﻿' (C) Copyright 2002-2009 by Autodesk, Inc. 
'
' Permission to use, copy, modify, and distribute this software in
' object code form for any purpose and without fee is hereby granted, 
' provided that the above copyright notice appears in all copies and 
' that both that copyright notice and the limited warranty and
' restricted rights notice below appear in all supporting 
' documentation.
'
' AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
' AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
' MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
' DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
' UNINTERRUPTED OR ERROR FREE.
'
' Use, duplication, or disclosure by the U.S. Government is subject to 
' restrictions set forth in FAR 52.227-19 (Commercial Computer
' Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
' (Rights in Technical Data and Computer Software), as applicable.
'
Imports System
Imports Autodesk.AutoCAD.Runtime
Imports Autodesk.AutoCAD.ApplicationServices
Imports Autodesk.AutoCAD.DatabaseServices
Imports Autodesk.AutoCAD.Geometry
Imports Autodesk.AutoCAD.EditorInput
Imports Autodesk.AutoCAD.GraphicsInterface

' This line is not mandatory, but improves loading performances
<Assembly: CommandClass(GetType(EEDOverrules_VB.MyCommands))> 

Namespace EEDOverrules_VB


#Region "Commands"
    Public Class MyCommands

        Private Shared mXrecOverrule As XrecordDrawOverrule
        Private Shared mXdataOverrule As XdataDrawOverrule
        Const dictName As String = "SGP_Data"

        Public Shared ReadOnly Property DictionaryName() As String
            Get
                Return dictName
            End Get
        End Property


        <CommandMethod("ADDXDATA")> _
        Public Shared Sub AddXdata()

            'If this is the first time our command has run, we create and register our overrule
            If mXdataOverrule Is Nothing Then
                mXdataOverrule = New XdataDrawOverrule
                Overrule.AddOverrule(RXClass.GetClass(GetType(Line)), mXdataOverrule, False)
                mXdataOverrule.SetXDataFilter(DictionaryName)
            End If
            'Ask user to select a line and enter the text to add to it as Xdata
            Dim db As Database = Application.DocumentManager.MdiActiveDocument.Database
            Dim ed As Editor = Application.DocumentManager.MdiActiveDocument.Editor
            Dim opts As New PromptEntityOptions(vbLf & "Select a line to add XDATA to: ")
            opts.SetRejectMessage(vbLf & "That's not a Line!" & vbLf)
            opts.AddAllowedClass(GetType(Line), True)
            Dim res As PromptEntityResult = ed.GetEntity(opts)

            If res.Status <> PromptStatus.OK Then Exit Sub

            Dim opts2 As New PromptStringOptions(vbLf & "Enter text to display: ")
            opts2.AllowSpaces = True
            opts2.DefaultValue = "end"

            Dim res2 As PromptResult = ed.GetString(opts2)

            If res2.Status <> PromptStatus.OK Then Exit Sub

            'If we've got to here, then a line was selected, and a string was entered.
            'Now we add the string to the line as Xdata.
            Using trans As Transaction = db.TransactionManager.StartTransaction
                'First we register the RegAppId (if its not there already)
                Dim appTbl As RegAppTable = trans.GetObject(db.RegAppTableId, OpenMode.ForRead)
                If appTbl.Has(DictionaryName) = False Then
                    Dim appTblRec As New RegAppTableRecord
                    appTbl.UpgradeOpen()
                    appTblRec.Name = DictionaryName
                    appTbl.Add(appTblRec)
                    trans.AddNewlyCreatedDBObject(appTblRec, True)
                End If

                'Then we add the Xdata
                'NOTE: This is very lazy programming - the following lines will remove any other application's Xdata from the line.
                Dim myLine As Line = trans.GetObject(res.ObjectId, OpenMode.ForWrite)
                myLine.XData = New ResultBuffer(New TypedValue(DxfCode.ExtendedDataRegAppName, DictionaryName), _
                                                New TypedValue(DxfCode.ExtendedDataAsciiString, res2.StringResult))

                trans.Commit()
            End Using

        End Sub

        <CommandMethod("ADDXRECORD")> _
        Public Shared Sub AddXrecord()

            'If this is the first time our command has run, we create and register our overrule
            If mXrecOverrule Is Nothing Then
                mXrecOverrule = New XrecordDrawOverrule
                Overrule.AddOverrule(RXClass.GetClass(GetType(Line)), mXrecOverrule, False)
                mXrecOverrule.SetExtensionDictionaryEntryFilter(dictName)
            End If
            'Ask user to select a line and enter the text to add to it as an Xrecord in the line's extension dictionary
            Dim db As Database = Application.DocumentManager.MdiActiveDocument.Database
            Dim ed As Editor = Application.DocumentManager.MdiActiveDocument.Editor
            Dim opts As New PromptEntityOptions(vbLf & "Select a line to add XRECORD to: ")
            opts.SetRejectMessage(vbLf & "That's not a Line!" & vbLf)
            opts.AddAllowedClass(GetType(Line), True)
            Dim res As PromptEntityResult = ed.GetEntity(opts)

            If res.Status <> PromptStatus.OK Then Exit Sub

            Dim opts2 As New PromptStringOptions(vbLf & "Enter text to display: ")
            opts2.AllowSpaces = True
            opts2.DefaultValue = "start"

            Dim res2 As PromptResult = ed.GetString(opts2)

            If res2.Status <> PromptStatus.OK Then Exit Sub

            'If we've got to here, then a line was selected, and a string was entered.
            'Now we add the string to the line in an xrecord in the line's extension dictionary

            Using trans As Transaction = db.TransactionManager.StartTransaction
                Dim myLine As Line = trans.GetObject(res.ObjectId, OpenMode.ForRead)
                Dim xDict As DBDictionary
                If myLine.ExtensionDictionary = ObjectId.Null Then
                    myLine.UpgradeOpen()
                    myLine.CreateExtensionDictionary()
                End If
                xDict = trans.GetObject(myLine.ExtensionDictionary, OpenMode.ForWrite)
                'Remove old data (if there is any)
                'NOTE: We're only removng our old data. Any other Xrecords or DBDictionaries will remain.
                If xDict.Contains(DictionaryName) Then
                    xDict.Remove(xDict.GetAt(DictionaryName))
                End If
                'Add new data
                Dim xRec As New Xrecord
                xRec.Data = New ResultBuffer(New TypedValue(DxfCode.ExtendedDataAsciiString, res2.StringResult))
                xDict.SetAt(DictionaryName, xRec)
                trans.AddNewlyCreatedDBObject(xRec, True)
                trans.Commit()
            End Using

        End Sub


        <CommandMethod("TOGGLEOVERRULE")> _
        Public Shared Sub ToggleOverrule()

            'Toggle value of Overruling, then regen graphics to show changes.
            Overrule.Overruling = Not Overrule.Overruling
            Application.DocumentManager.MdiActiveDocument.Editor.Regen()

        End Sub

    End Class
#End Region



#Region "Xrecord Overrule"
    Public Class XrecordDrawOverrule
        Inherits DrawableOverrule

        'Return false from WorldDraw to ensure ViewportDraw is called
        Public Overrides Function WorldDraw(ByVal drawable As Autodesk.AutoCAD.GraphicsInterface.Drawable, ByVal wd As Autodesk.AutoCAD.GraphicsInterface.WorldDraw) As Boolean

            MyBase.WorldDraw(drawable, wd)
            Return False

        End Function


        Public Overrides Sub ViewportDraw(ByVal drawable As Autodesk.AutoCAD.GraphicsInterface.Drawable, ByVal vd As Autodesk.AutoCAD.GraphicsInterface.ViewportDraw)

            Dim myLine As Line = drawable
            If Not myLine.Database Is Nothing Then

                Dim displayText As String
                Using trans As OpenCloseTransaction = myLine.Database.TransactionManager.StartOpenCloseTransaction
                    'Retrieve EED
                    'NOTE: Because we know this overrule will only be called for a line that has our entry in its extension dictionary
                    '   we know the data already exists - so we don't have to test if its there before we access it.
                    Dim xDict As DBDictionary = trans.GetObject(myLine.ExtensionDictionary, OpenMode.ForRead)
                    Dim xRec As Xrecord = trans.GetObject(xDict.GetAt(MyCommands.DictionaryName), OpenMode.ForRead)
                    displayText = xRec.Data.AsArray(0).Value
                End Using
                'Write some text in the default textstyle at the end point of the line
                'The text will be horizontal to the view.
                Dim txtStyle As New TextStyle
                Dim mat As Matrix3d = vd.Viewport.EyeToWorldTransform
                vd.Geometry.Text(myLine.StartPoint, vd.Viewport.ViewDirection, mat.CoordinateSystem3d.Xaxis, displayText, True, txtStyle)
            End If
            MyBase.ViewportDraw(drawable, vd)

        End Sub

    End Class
#End Region


#Region "Xdata Overrule"
    Public Class XdataDrawOverrule
        Inherits DrawableOverrule

        'Return false from WorldDraw to ensure ViewportDraw is called
        Public Overrides Function WorldDraw(ByVal drawable As Autodesk.AutoCAD.GraphicsInterface.Drawable, ByVal wd As Autodesk.AutoCAD.GraphicsInterface.WorldDraw) As Boolean

            MyBase.WorldDraw(drawable, wd)
            Return False

        End Function

        Public Overrides Sub ViewportDraw(ByVal drawable As Autodesk.AutoCAD.GraphicsInterface.Drawable, ByVal vd As Autodesk.AutoCAD.GraphicsInterface.ViewportDraw)

            Dim myLine As Line = drawable
            If Not myLine.Database Is Nothing Then
                Dim displayText As String
                Using trans As OpenCloseTransaction = myLine.Database.TransactionManager.StartOpenCloseTransaction
                    'Retrieve EED
                    'NOTE: Because we know this overrule will only be called for a line that has our entry in its Xdata
                    '      we know the data already exists - so we don't have to test if its there before we access it.
                    displayText = myLine.GetXDataForApplication(MyCommands.DictionaryName).AsArray(1).Value
                End Using
                'Write some text in the default textstyle at the end point of the line
                'The text will be horizontal to the view.
                Dim txtStyle As New TextStyle
                Dim mat As Matrix3d = vd.Viewport.EyeToWorldTransform
                vd.Geometry.Text(myLine.EndPoint, mat.CoordinateSystem3d.Zaxis, mat.CoordinateSystem3d.Xaxis, displayText, True, txtStyle)
            End If
            MyBase.ViewportDraw(drawable, vd)

        End Sub

    End Class
#End Region

End Namespace