﻿' (C) Copyright 2002-2009 by Autodesk, Inc. 
'
' Permission to use, copy, modify, and distribute this software in
' object code form for any purpose and without fee is hereby granted, 
' provided that the above copyright notice appears in all copies and 
' that both that copyright notice and the limited warranty and
' restricted rights notice below appear in all supporting 
' documentation.
'
' AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
' AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
' MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
' DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
' UNINTERRUPTED OR ERROR FREE.
'
' Use, duplication, or disclosure by the U.S. Government is subject to 
' restrictions set forth in FAR 52.227-19 (Commercial Computer
' Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
' (Rights in Technical Data and Computer Software), as applicable.
'
Imports System
Imports Autodesk.AutoCAD.Runtime
Imports Autodesk.AutoCAD.ApplicationServices
Imports Autodesk.AutoCAD.DatabaseServices
Imports Autodesk.AutoCAD.Geometry
Imports Autodesk.AutoCAD.EditorInput

' This line is not mandatory, but improves loading performances
<Assembly: CommandClass(GetType(ObjectIdFilterOverrule_VB.MyCommands))> 

Namespace ObjectIdFilterOverrule_VB

#Region "MyCommands"
    Public Class MyCommands

        'Stores ObjectIds of entities we want to overrule
        Private Shared mObjIds As New ObjectIdCollection
        'Stores global instance of our Overrule class
        Private Shared mOverrule As MyObjectOverrule

        'If a document is destroyed, we need to remove any ObjectIds in that document's database from our list
        Private Shared WithEvents mDocManager As DocumentCollection
        Private Shared Sub mDocManager_DocumentToBeDestroyed(ByVal sender As Object, ByVal e As DocumentCollectionEventArgs) Handles mDocManager.DocumentToBeDestroyed
            For i As Integer = mObjIds.Count - 1 To 0 Step -1
                If mObjIds.Item(i).Database = e.Document.Database Then
                    mObjIds.RemoveAt(i)
                End If
            Next
        End Sub


        'Asks user to select an entity. If entity is not already in overrule list, then it is added.
        'If already in list, it is removed
        <CommandMethod("TOGGLELOCK")> _
        Public Shared Sub ToggleLock()
            Dim ed As Editor = Application.DocumentManager.MdiActiveDocument.Editor
            Dim opts As New PromptEntityOptions("Select an entity to toggle overrule: ")
            Dim res As PromptEntityResult = ed.GetEntity(opts)

            If res.Status <> PromptStatus.OK Then Exit Sub

            'If this is the first time we've run this command, or if we called ClearLocks, we create and register our overrule.
            'Also initialize our DocumentManager events so we can remove entities from our list if we close the doc they are in,
            ' and turn overruling on.
            If mOverrule Is Nothing Then
                mOverrule = New MyObjectOverrule
                Overrule.AddOverrule(RXClass.GetClass(GetType(Entity)), mOverrule, True)
                mDocManager = Application.DocumentManager
                Overrule.Overruling = True
            End If
            'Add or remove selected Objectid from the collection
            If mObjIds.Contains(res.ObjectId) Then
                mObjIds.Remove(res.ObjectId)
            Else
                mObjIds.Add(res.ObjectId)
            End If
            'Send our ObjectIdCollection to the overrule
            mOverrule.SetIdFilter(IdColl2Array(mObjIds))
        End Sub

        Private Shared Function IdColl2Array(ByVal idcoll As ObjectIdCollection) As ObjectId()
            Dim myIdArray(idcoll.Count - 1) As ObjectId
            For i As Integer = 0 To idcoll.Count - 1
                myIdArray(i) = idcoll.Item(i)
            Next
            Return myIdArray
        End Function


        'Clears all overrules
        <CommandMethod("CLEARLOCKS")> _
        Public Shared Sub ClearLocks()
            If Not mOverrule Is Nothing Then
                Overrule.RemoveOverrule(RXClass.GetClass(GetType(Entity)), mOverrule)
                mOverrule = Nothing
                mDocManager = Nothing
                Overrule.Overruling = False
            End If
            mObjIds.Clear()


        End Sub


    End Class
#End Region

#Region "MyObjectOverrule - a custom ObjectOverrule"


    'Simple overrule just to stop an entity being erased.
    Public Class MyObjectOverrule
        Inherits ObjectOverrule

        Public Overrides Sub [Erase](ByVal dbObject As DBObject, ByVal erasing As Boolean)
            MyBase.[Erase](dbObject, erasing)
            'prevent the object from being deleted
            If erasing Then
                'This will cause your debugger to break here to report the exception because we're not handling it.
                'That's ok because the exception is intended for AutoCAD.
                Throw New Global.Autodesk.AutoCAD.Runtime.Exception(ErrorStatus.CannotBeErasedByCaller)
            End If

        End Sub

    End Class
#End Region

End Namespace