﻿' (C) Copyright 2010 by Microsoft 
'
Imports System
Imports Autodesk.AutoCAD.Runtime
Imports Autodesk.AutoCAD.ApplicationServices
Imports Autodesk.AutoCAD.DatabaseServices
Imports Autodesk.AutoCAD.Geometry
Imports Autodesk.AutoCAD.EditorInput
Imports Autodesk.AutoCAD.GraphicsInterface
Imports System.Collections.Generic

' This line is not mandatory, but improves loading performances
<Assembly: CommandClass(GetType(OverruleXdataCopyTest.MyCommands))> 

Namespace OverruleXdataCopyTest

#Region "Commands"

  Public Class MyCommands

    Private Shared mLayerId As ObjectId


    ' Adds a custom 'leader line' to a block reference selected by the user
    <CommandMethod("ADDLINE")> _
       Public Shared Sub TEST()
      Dim db As Database = Application.DocumentManager.MdiActiveDocument.Database
      Dim ed As Editor = Application.DocumentManager.MdiActiveDocument.Editor

      ' Prompt user to select a block reference and start/end point for the leader line

      Dim opts As New PromptEntityOptions(vbLf & "Select Block Reference to overrule: ")
      opts.SetRejectMessage(vbLf & "That isn't a Block Reference!" & vbLf)
      opts.AddAllowedClass(GetType(BlockReference), True)
      Dim res As PromptEntityResult = ed.GetEntity(opts)

      If res.Status <> PromptStatus.OK Then Exit Sub

      Dim ptOpts As New PromptPointOptions("Select leader start point:")
      Dim ptRes As PromptPointResult = ed.GetPoint(ptOpts)

      If ptRes.Status <> PromptStatus.OK Then Exit Sub

      Dim ptOpts2 As New PromptPointOptions("Select leader end point:")
      ptOpts2.BasePoint = ptRes.Value
      ptOpts2.UseBasePoint = True
      Dim ptRes2 As PromptPointResult = ed.GetPoint(ptOpts2)

      If ptRes.Status <> PromptStatus.OK Then Exit Sub

      ' If user didn't cancel command then we add leader line start/end points as xdata to the blockreference 

      ' First create our RegAppId (if it doesn't already exist)
      Using trans As Transaction = db.TransactionManager.StartTransaction
        Dim appTbl As RegAppTable = trans.GetObject(db.RegAppTableId, OpenMode.ForRead)
        If appTbl.Has(SGPOverruleBag.DictName) = False Then
          Dim appTblRec As New RegAppTableRecord
          appTbl.UpgradeOpen()
          appTblRec.Name = SGPOverruleBag.DictName
          appTbl.Add(appTblRec)
          trans.AddNewlyCreatedDBObject(appTblRec, True)
        End If

        trans.Commit()
      End Using

      ' Then we add the xdata to the block ref
      SGPOverruleBag.setData(res.ObjectId, New ResultBuffer(New TypedValue(DxfCode.ExtendedDataWorldXCoordinate, ptRes.Value), New TypedValue(DxfCode.ExtendedDataWorldXCoordinate, ptRes2.Value)))

    End Sub

  End Class

#End Region


  Public Class SGPOverruleBag
    Implements IDisposable


#Region "Helper functions"


    Private Shared mRegisterClass As RXClass = RXClass.GetClass(GetType(BlockReference))
    Private mOverrules As System.Collections.Generic.List(Of Overrule)

    Private Const mDictName = "SGP_DATA"
    Friend Shared ReadOnly Property DictName()
      Get
        Return mDictName
      End Get
    End Property

    'Two versions of New - if no RXClass is passed in then we instantiate the overrules, but don't regieter them
    Public Sub New()
      subNew()
    End Sub

    Public Sub New(ByVal regClass As RXClass)
      subNew()
      RegisterClass = regClass
      If Not regClass Is Nothing Then
        Register(RegisterClass)
      End If
    End Sub

    Private Sub subNew()
      MyOverrules = New System.Collections.Generic.List(Of Overrule)
      MyOverrules.Add(New SGPDrawOverrule)
      MyOverrules.Add(New SGPGripOverrule)
      MyOverrules.Add(New SGPOsnapOverrule)
      MyOverrules.Add(New SGPGeometryOverrule)
      MyOverrules.Add(New SGPTransformOverrule)
    End Sub

    ' Set the  class that overrules are rregistered for
    Friend Shared Property RegisterClass() As RXClass
      Get
        Return mRegisterClass
      End Get
      Set(ByVal value As RXClass)
        mRegisterClass = value
      End Set
    End Property

    ' List of overrules applied by DSOverruleBag
    Private Property MyOverrules() As System.Collections.Generic.List(Of Overrule)
      Get
        Return mOverrules
      End Get
      Set(ByVal value As System.Collections.Generic.List(Of Overrule))
        mOverrules = value
      End Set
    End Property

    ' Register and Unregister will normally be called from this class's New and Dispose methods.
    ' However, they may be called from another class to register the overrules against more than one class.
    Friend Sub Register(ByVal regClass As RXClass)
      If Not regClass Is Nothing Then
        For Each orule As Overrule In MyOverrules
          Overrule.AddOverrule(regClass, orule, False)
          orule.SetXDataFilter(DictName)
        Next
        Overrule.Overruling = True
      End If
    End Sub

    Friend Sub Register()
      Register(RegisterClass)
    End Sub

    Friend Sub Unregister(ByVal regClass As RXClass)
      If Not regClass Is Nothing Then
        For Each orule As Overrule In MyOverrules
          Overrule.RemoveOverrule(regClass, orule)
        Next
        Overrule.Overruling = False
      End If
    End Sub

    Friend Sub Unregister()
      Unregister(RegisterClass)
    End Sub


    ' Adds xdata to a DBObject
    ' obj is assumed to be already open for read
    Friend Shared Sub SetData(ByVal obj As DBObject, ByVal resbuf As ResultBuffer)
      'Start with our regappid
      Dim rb As New ResultBuffer(New TypedValue(DxfCode.ExtendedDataRegAppName, DictName))
      'Then append our data to the resbuf we want to add as xdata
      For Each tv As TypedValue In resbuf
        rb.Add(tv)
      Next
      'Add the xdata
      obj.XData = rb

    End Sub

    ' Adds xdata to a DBObject with ObjectId objId
    Friend Shared Sub SetData(ByVal objId As ObjectId, ByVal resbuf As ResultBuffer)
      Dim db As Database = objId.Database
      Using trans = db.TransactionManager.StartOpenCloseTransaction
        Dim obj As DBObject = trans.GetObject(objId, OpenMode.ForWrite)
        SetData(obj, resbuf)
        trans.Commit()
      End Using
    End Sub

    ' Retrieves xdata from DBObject with ObjectId objId
    Friend Shared Function GetData(ByVal objId As ObjectId) As ResultBuffer
      Dim db As Database = objId.Database
      Dim resbuf As ResultBuffer
      Using trans = db.TransactionManager.StartOpenCloseTransaction
        Dim obj As DBObject = trans.GetObject(objId, OpenMode.ForRead)
        resbuf = GetData(obj)
        trans.Commit()
      End Using
      Return resbuf
    End Function

    ' Retrieves xdata from DBObject 
    '  obj is assumed to be open for read
    Friend Shared Function GetData(ByVal obj As DBObject) As ResultBuffer

      Dim rb As ResultBuffer = obj.XData
      Dim arr() As TypedValue = rb.AsArray
      'Find our xdata. There could be xdata from other apps.
      Dim index As Integer = Array.FindIndex(arr, AddressOf MyXdata)
      If index = -1 Then Return Nothing
      Dim index2 As Integer = Array.FindIndex(arr, index + 1, AddressOf AnyXdata)
      If index2 = -1 Then index2 = arr.Count
      Dim length As Integer = index2 - index - 1
      Dim arr2(length - 1) As TypedValue
      Array.Copy(arr, index + 1, arr2, 0, length)
      Dim rb2 As New ResultBuffer
      For Each tv As TypedValue In arr2
        rb2.Add(tv)
      Next
      Return rb2
    End Function

    'Helper function used by Array.FindIndex
    Private Shared Function MyXdata(ByVal tv As TypedValue) As Boolean
      If tv.TypeCode = DxfCode.ExtendedDataRegAppName Then
        If tv.Value = DictName Then
          Return True
        End If
      End If
      Return False
    End Function

    'Helper function used by Array.FindIndex
    Private Shared Function AnyXdata(ByVal tv As TypedValue) As Boolean
      If tv.TypeCode = DxfCode.ExtendedDataRegAppName Then
        Return True
      Else
        Return False
      End If
    End Function




    Private disposedValue As Boolean = False        ' To detect redundant calls

    ' IDisposable
    Protected Overridable Overloads Sub Dispose(ByVal disposing As Boolean)
      If Not Me.disposedValue Then
        If disposing Then
          ' TODO: free other state (managed objects).
          Unregister(RegisterClass)
          For Each orule As Overrule In MyOverrules
            orule.Dispose()
          Next
          mOverrules = Nothing
        End If
        ' TODO: free your own state (unmanaged objects).
        ' TODO: set large fields to null.
      End If
      Me.disposedValue = True
    End Sub


#Region " IDisposable Support "
    ' This code added by Visual Basic to correctly implement the disposable pattern.
    Public Overloads Sub Dispose() Implements IDisposable.Dispose
      ' Do not change this code.  Put cleanup code in Dispose(ByVal disposing As Boolean) above.
      Dispose(True)
      GC.SuppressFinalize(Me)
    End Sub

    Protected Overrides Sub Finalize()
      Dispose(False)
      MyBase.Finalize()
    End Sub

#End Region


#End Region

#Region "Overrules"
    Private Class SGPGeometryOverrule
      Inherits GeometryOverrule

      ' Note: This is not needed (or evcen called) in 2D wireframe mode, but is in 3D wireframe.
      Public Overrides Function GetGeomExtents(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity) As Autodesk.AutoCAD.DatabaseServices.Extents3d
        Dim ext As Extents3d = MyBase.GetGeomExtents(entity)
        'Add extents of leader line to extents of MyBase
        Dim rb As ResultBuffer = SGPOverruleBag.GetData(entity)
        If Not rb Is Nothing Then
          ext.AddPoint(rb.AsArray(0).Value)
          ext.AddPoint(rb.AsArray(1).Value)
        End If
        'Return total extents
        Return ext
      End Function

    End Class



    Private Class SGPGripOverrule
      Inherits GripOverrule

      ' Define our custom grip class
      Public Class MyGrip
        Inherits GripData
        Private mGripNum As Integer

        ' Ordinal is used to distinguish grips of the same class (in this case leader line start and end point grips.
        Public Property Ordinal() As Integer
          Get
            Return mGripNum
          End Get
          Set(ByVal value As Integer)
            mGripNum = value
          End Set
        End Property

        'Call this to tell the grip to move itself
        Public Sub Move(ByVal vec As Vector3d)
          GripPoint = GripPoint + vec
        End Sub

        'Grip draws itself
        Public Overrides Function ViewportDraw(ByVal worldDraw As Autodesk.AutoCAD.GraphicsInterface.ViewportDraw, ByVal entityId As Autodesk.AutoCAD.DatabaseServices.ObjectId, ByVal type As Autodesk.AutoCAD.DatabaseServices.GripData.DrawType, ByVal imageGripPoint As Autodesk.AutoCAD.Geometry.Point3d?, ByVal gripSizeInPixels As Integer) As Boolean
          Dim unit As Point2d = worldDraw.Viewport.GetNumPixelsInUnitSquare(GripPoint)
          worldDraw.Geometry.Circle(GripPoint, gripSizeInPixels / unit.X, worldDraw.Viewport.ViewDirection)
          Return True
        End Function

        Public Overrides Function GetTooltip() As String
          Return "My Custom Grip"
        End Function

      End Class


      'Array to hold our 2 grips
      Dim mGripData(1) As GripData

      ' Add our two custom grips to GripDataCollection and then call base function to also add grips from block ref
      Public Overrides Sub GetGripPoints(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal grips As Autodesk.AutoCAD.DatabaseServices.GripDataCollection, ByVal curViewUnitSize As Double, ByVal gripSize As Integer, ByVal curViewDir As Autodesk.AutoCAD.Geometry.Vector3d, ByVal bitFlags As Autodesk.AutoCAD.DatabaseServices.GetGripPointsFlags)
        Dim grip1 As New MyGrip
        Dim rb As ResultBuffer = SGPOverruleBag.GetData(entity)
        grip1.Ordinal = 1
        grip1.GripPoint = rb.AsArray(0).Value
        mGripData(0) = grip1
        grips.Add(grip1)

        Dim grip2 As New MyGrip
        grip2.Ordinal = 2
        grip2.GripPoint = rb.AsArray(1).Value
        mGripData(0) = grip2
        grips.Add(grip2)

        MyBase.GetGripPoints(entity, grips, curViewUnitSize, gripSize, curViewDir, bitFlags)
      End Sub


      Public Overrides Sub MoveGripPointsAt(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal grips As Autodesk.AutoCAD.DatabaseServices.GripDataCollection, ByVal offset As Autodesk.AutoCAD.Geometry.Vector3d, ByVal bitFlags As Autodesk.AutoCAD.DatabaseServices.MoveGripPointsFlags)

        Dim rb As ResultBuffer = SGPOverruleBag.GetData(entity)
        Dim pts(1) As Point3d
        pts(0) = rb.AsArray(0).Value
        pts(1) = rb.AsArray(1).Value

        ' If moved grip is one of ours, then we want to update our graphics accordingly - i.e. we modify the xdata for the leader line start or end point
        For Each g As GripData In grips
          If TypeOf g Is MyGrip Then
            Dim mat As Matrix3d = Matrix3d.Displacement(offset)
            Dim grip As MyGrip = g
            pts(grip.Ordinal - 1) = grip.GripPoint + offset
            SGPOverruleBag.SetData(entity, New ResultBuffer(New TypedValue(DxfCode.ExtendedDataWorldXCoordinate, pts(0)), New TypedValue(DxfCode.ExtendedDataWorldXCoordinate, pts(1))))
          End If
        Next


        MyBase.MoveGripPointsAt(entity, grips, offset, bitFlags)
      End Sub

    End Class


    Private Class SGPOsnapOverrule
      Inherits OsnapOverrule



      Public Overrides Sub GetObjectSnapPoints(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal snapMode As Autodesk.AutoCAD.DatabaseServices.ObjectSnapModes, ByVal gsSelectionMark As System.IntPtr, ByVal pickPoint As Autodesk.AutoCAD.Geometry.Point3d, ByVal lastPoint As Autodesk.AutoCAD.Geometry.Point3d, ByVal viewTransform As Autodesk.AutoCAD.Geometry.Matrix3d, ByVal snapPoints As Autodesk.AutoCAD.Geometry.Point3dCollection, ByVal geometryIds As Autodesk.AutoCAD.Geometry.IntegerCollection)

        ' Use xdata to construct a temporary line to return the snap points we need
        Dim rb As ResultBuffer = SGPOverruleBag.GetData(entity)
        If Not rb Is Nothing Then
          Using theLine As New Line(rb.AsArray(0).Value, rb.AsArray(1).Value)
            theLine.GetObjectSnapPoints(snapMode, gsSelectionMark, pickPoint, lastPoint, viewTransform, snapPoints, geometryIds)
          End Using
        End If

        MyBase.GetObjectSnapPoints(entity, snapMode, gsSelectionMark, pickPoint, lastPoint, viewTransform, snapPoints, geometryIds)

      End Sub

      Public Overrides Sub GetObjectSnapPoints(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal snapMode As Autodesk.AutoCAD.DatabaseServices.ObjectSnapModes, ByVal gsSelectionMark As System.IntPtr, ByVal pickPoint As Autodesk.AutoCAD.Geometry.Point3d, ByVal lastPoint As Autodesk.AutoCAD.Geometry.Point3d, ByVal viewTransform As Autodesk.AutoCAD.Geometry.Matrix3d, ByVal snapPoints As Autodesk.AutoCAD.Geometry.Point3dCollection, ByVal geometryIds As Autodesk.AutoCAD.Geometry.IntegerCollection, ByVal insertionMat As Autodesk.AutoCAD.Geometry.Matrix3d)
        Dim rb As ResultBuffer = SGPOverruleBag.GetData(entity)

        ' Use xdata to construct a temporary line to return the snap points we need

        If Not rb Is Nothing Then
          Using theLine As New Line(rb.AsArray(0).Value, rb.AsArray(1).Value)
            theLine.GetObjectSnapPoints(snapMode, gsSelectionMark, pickPoint, lastPoint, viewTransform, snapPoints, geometryIds, insertionMat)
          End Using
        End If

        MyBase.GetObjectSnapPoints(entity, snapMode, gsSelectionMark, pickPoint, lastPoint, viewTransform, snapPoints, geometryIds, insertionMat)

      End Sub

    End Class




    Private Class SGPDrawOverrule
      Inherits DrawableOverrule


      Public Overrides Function WorldDraw(ByVal drawable As Autodesk.AutoCAD.GraphicsInterface.Drawable, ByVal wd As Autodesk.AutoCAD.GraphicsInterface.WorldDraw) As Boolean

        Dim myBR As Autodesk.AutoCAD.DatabaseServices.BlockReference = drawable

        ' Retrieve xdata and use it to draw our leader line
        Using rb As ResultBuffer = SGPOverruleBag.GetData(myBR)
          If Not rb Is Nothing Then
            Dim pts As New Point3dCollection
            pts.Add(rb.AsArray(0).Value)
            pts.Add(rb.AsArray(1).Value)
            wd.Geometry.Polyline(pts, myBR.Normal, New IntPtr(4321))
          End If
        End Using
        Return MyBase.WorldDraw(drawable, wd)

      End Function

    End Class


    Private Class SGPTransformOverrule
      Inherits TransformOverrule

      Public Overrides Sub Explode(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal entitySet As Autodesk.AutoCAD.DatabaseServices.DBObjectCollection)

        ' Let block ref explode itself
        MyBase.Explode(entity, entitySet)

        ' Then create new Line from our xdata and add it to entitySet
        Dim rb As ResultBuffer = SGPOverruleBag.GetData(entity)
        If Not rb Is Nothing Then
          Dim theLine As New Line(rb.AsArray(0).Value, rb.AsArray(1).Value)
          entitySet.Add(theLine)
        End If
      End Sub

    End Class

#End Region



  End Class


End Namespace