﻿' (C) Copyright 2010 by Autodesk, Inc. 
'
' Permission to use, copy, modify, and distribute this software in
' object code form for any purpose and without fee is hereby granted,
' provided that the above copyright notice appears in all copies and
' that both that copyright notice and the limited warranty and
' restricted rights notice below appear in all supporting
' documentation.
'
' AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
' AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
' MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
' DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
' UNINTERRUPTED OR ERROR FREE.
'
' Use, duplication, or disclosure by the U.S. Government is subject
' to restrictions set forth in FAR 52.227-19 (Commercial Computer
' Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
' (Rights in Technical Data and Computer Software), as applicable.
'
Imports Autodesk.AutoCAD.Runtime
Imports Autodesk.AutoCAD.ApplicationServices
Imports Autodesk.AutoCAD.DatabaseServices

Namespace DigSigStamp

  ' This class takes care of initialization and termination of the 
  '  addin. It also deals with setting/unsetting the DBMODProxy info
  '  required  to workaround PLOT and PUBLISH setting DBMOD > 0.
  '  See later comments about that.

  Public Class MyPlugin
    Implements IExtensionApplication

    Private Shared mOruleClassType As Type = GetType(BlockReference)
    Private Shared WithEvents docs As DocumentCollection
    Private Shared WithEvents thisDoc As Document
    Dim mOverrules As DSOverruleBag

    Friend Shared ReadOnly Property OverruleClassType() As Type
      Get
        Return mOruleClassType
      End Get
    End Property

    Public Sub Initialize() _
    Implements IExtensionApplication.Initialize

      'Register for demand loading

      DemandLoading.RegistryUpdate.RegisterForDemandLoading()

      ' Define the AutoCAD object class you want to overrule here

      mOruleClassType = GetType(BlockReference)
      If mOverrules Is Nothing Then

        ' DSOverruleBag's constructor handles registering the
        '  Overrule()

        mOverrules = New  _
        DSOverruleBag(RXClass.GetClass(OverruleClassType))
        docs = Application.DocumentManager
        thisDoc = docs.MdiActiveDocument

      End If
    End Sub

    Public Sub Terminate() Implements IExtensionApplication.Terminate

      ' Do plug-in application clean up here
      ' Terminate is only called when AutoCAD is quitting,
      '  so no need to explicitly dispose of or unregister Overrules.

      docs = Nothing
      thisDoc = Nothing
    End Sub

    ' This is required  to make sure out thisDoc events are always 
    '  related to the active document.

    Private Shared Sub docs_DocumentActivated( _
      ByVal sender As Object, _
      ByVal e As DocumentCollectionEventArgs) _
      Handles docs.DocumentActivated

      thisDoc = e.Document
    End Sub




    ' If the user invokes the PLOT dialog, it will cause the drawing
    '  to be flagged as dirty (DBMOD > 0). This causes our DBMOD
    '  check when deciding how to draw block reference overrule to
    '  always report that the drawing is dirty. That means just
    '  checking DBMOD causes every plot to be flagged as dirty.
    ' We work around that by checking if DBMOD is zero as the PLOT 
    '  command is starting. We add userdata to the document if it
    '  was, and use that to tell us how to draw our block reference
    '  during a plot.
    ' Note: Our userdata attached to a drawing means the drawing was
    '  clean just before plotting. No userdata means it was dirty.

    Private Shared Sub thisDoc_CommandWillStart(ByVal sender As _
      Object, ByVal e As CommandEventArgs) _
      Handles thisDoc.CommandWillStart

      Dim DBMOD As Short = _
        DirectCast(Application.GetSystemVariable("DBMOD"), Short)
      If e.GlobalCommandName = "PLOT" _
        Or e.GlobalCommandName = "PUBLISH" Then
        If DBMOD = 0 Then
          thisDoc.UserData.Add(myCommands.DictName, 0)
        ElseIf thisDoc.UserData.Contains(myCommands.DictName) Then
          thisDoc.UserData.Remove(myCommands.DictName)
        End If
      End If
    End Sub

    ' Clearing our DBMODProxy userdata when any command ends -
    '  just to be safe.
    Private Shared Sub thisDoc_CommandCancelled(ByVal sender As _
      Object, ByVal e As CommandEventArgs) _
      Handles thisDoc.CommandCancelled

      If thisDoc.UserData.Contains(myCommands.DictName) Then
        thisDoc.UserData.Remove(myCommands.DictName)
      End If
    End Sub

    Private Shared Sub thisDoc_CommandEnded(ByVal sender As Object, _
      ByVal e As CommandEventArgs) Handles thisDoc.CommandEnded
      If thisDoc.UserData.Contains(myCommands.DictName) Then
        thisDoc.UserData.Remove(myCommands.DictName)
      End If
    End Sub

    Private Shared Sub thisDoc_CommandFailed( _
      ByVal sender As Object, ByVal e As CommandEventArgs) _
      Handles thisDoc.CommandFailed

      If thisDoc.UserData.Contains(MyCommands.DictName) Then
        thisDoc.UserData.Remove(MyCommands.DictName)
      End If
    End Sub

  End Class

End Namespace