﻿' (C) Copyright 2010 by Autodesk, Inc. 
'
' Permission to use, copy, modify, and distribute this software in
' object code form for any purpose and without fee is hereby granted,
' provided that the above copyright notice appears in all copies and 
' that both that copyright notice and the limited warranty and
' restricted rights notice below appear in all supporting 
' documentation.
'
' AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
' AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
' MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
' DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
' UNINTERRUPTED OR ERROR FREE.
'
' Use, duplication, or disclosure by the U.S. Government is subject
' to restrictions set forth in FAR 52.227-19 (Commercial Computer
' Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
' (Rights in Technical Data and Computer Software), as applicable.
'
Imports System
Imports Autodesk.AutoCAD.Runtime
Imports Autodesk.AutoCAD.ApplicationServices
Imports Autodesk.AutoCAD.DatabaseServices
Imports Autodesk.AutoCAD.Geometry
Imports Autodesk.AutoCAD.EditorInput

' This line is not mandatory, but improves loading performance

<Assembly: CommandClass(GetType(DimensionPatrol.MyCommands))> 

Namespace DimensionPatrol

  ' Container class for our commands
  ' All methods and properties in this class are Shared, 
  '  so this class is never instantiated by AutoCAD.

  Public Class MyCommands

    ' Member variables to store values of DIMPATROL and
    ' DIMPATROLCOLOR

    Private Shared mDimPatrol As Short = 0
    Private Shared mDimPatrolColor As Short = 1

    ' This property is also used by MyPlugin and
    ' DimPatrolDrawOverrule - hence use of 'Friend'

    Friend Shared Property DimPatrolVal() As Short

      Get
        Return mDimPatrol
      End Get

      Private Set(ByVal value As Short)

        mDimPatrol = value

        ' Tell MyPlugin that value of DimPatrol has changed

        MyPlugin.DimPatVar_Changed()

      End Set

    End Property

    ' This property is also used by MyPlugin and
    ' DimPatrolDrawOverrule - hence use of 'Friend'

    Friend Shared Property DimPatrolColVal() As Short

      Get
        Return mDimPatrolColor
      End Get

      Private Set(ByVal value As Short)

        mDimPatrolColor = value

        ' Tell MyPlugin that value of DimPatrolColor has changed

        MyPlugin.DimPatColorVar_Changed()

      End Set

    End Property


    ' Turn dimension patrol on (1) or off (0)

    <CommandMethod("ADNPLUGINS", "DIMPATROL", CommandFlags.Modal)> _
    Public Shared Sub DIMPATROL()

      Try

        ' Prompt for new value

        Dim ed As Editor
        ed = Application.DocumentManager.MdiActiveDocument.Editor
        Dim opts As New  _
          PromptIntegerOptions("Enter new value for DIMPATROL")
        opts.DefaultValue = DimPatrolVal
        opts.LowerLimit = 0
        opts.UpperLimit = 1
        Dim res As PromptIntegerResult = ed.GetInteger(opts)

        ' Only update stored value if user has entered a different
        '  value()

        If (res.Status = PromptStatus.OK) _
          And (res.Value <> DimPatrolVal) Then

          DimPatrolVal = res.Value

        End If

      Catch ex As Autodesk.AutoCAD.Runtime.Exception
        Application.DocumentManager.MdiActiveDocument. _
          Editor.WriteMessage( _
            vbLf & "Error changing DIMPATROL value: " & _
            ex.ErrorStatus.ToString & vbLf)
      End Try

    End Sub


    ' Allow user to set highlight color for manually edited
    ' dimensions. Color index can be set within range 1-7.

    <CommandMethod("ADNPLUGINS", "DIMPATROLCOLOR", _
                   CommandFlags.Modal)> _
    Public Shared Sub DIMPATROLCOLOR()

      Try

        ' Prompt for new value

        Dim ed As Editor
        ed = Application.DocumentManager.MdiActiveDocument.Editor
        Dim opts As New  _
          PromptIntegerOptions("Enter new value for DIMPATROLCOLOR")
        opts.DefaultValue = DimPatrolColVal
        opts.LowerLimit = 1
        opts.UpperLimit = 7
        Dim res As PromptIntegerResult = ed.GetInteger(opts)

        ' Only update stored value if user has entered a different 
        '  value

        If (res.Status = PromptStatus.OK) _
          And (res.Value <> DimPatrolColVal) Then

          DimPatrolColVal = res.Value

        End If

      Catch ex As Autodesk.AutoCAD.Runtime.Exception
        Application.DocumentManager.MdiActiveDocument.Editor. _
          WriteMessage( _
            vbLf & "Error changing DIMPATROLCOLOR value: " _
            & ex.ErrorStatus.ToString & vbLf)
      End Try

    End Sub


    'Remove demand load settings from registry

    <CommandMethod("ADNPLUGINS", "REMOVEDP", CommandFlags.Modal)> _
    Public Shared Sub RemoveDP()

      DemandLoading.RegistryUpdate.UnregisterForDemandLoading()

      Dim ed As Editor = _
        Autodesk.AutoCAD.ApplicationServices.Application _
        .DocumentManager.MdiActiveDocument.Editor()
      ed.WriteMessage(vbCr + _
        "Dimension Patrol will not be loaded" _
        + " automatically in future editing sessions.")

    End Sub

  End Class

End Namespace