﻿
Imports System
Imports Autodesk.AutoCAD.Runtime
Imports Autodesk.AutoCAD.ApplicationServices
Imports Autodesk.AutoCAD.DatabaseServices
Imports Autodesk.AutoCAD.Geometry
Imports Autodesk.AutoCAD.EditorInput
Imports Autodesk.AutoCAD.GraphicsInterface
Imports System.Collections.Generic
Imports Autodesk.AutoCAD.Colors

' This line is not mandatory, but improves loading performances
<Assembly: CommandClass(GetType(SGP_Network.MyCommands))> 

Namespace SGP_Network


#Region "Commands"

  Public Class MyCommands


    ' Add network link between two nodes (circles)
    <CommandMethod("ADDLINK")> _
       Public Shared Sub AddLink()
      Dim db As Database = Application.DocumentManager.MdiActiveDocument.Database
      Dim ed As Editor = Application.DocumentManager.MdiActiveDocument.Editor
      ' Select first node (circle)
      Dim opts As New PromptEntityOptions(vbLf & "Select start node (circle) for connection: ")
      opts.SetRejectMessage(vbLf & "That isn't a Circle!" & vbLf)
      opts.AddAllowedClass(GetType(Circle), True)
      Dim res As PromptEntityResult = ed.GetEntity(opts)
      If res.Status <> PromptStatus.OK Then Exit Sub
      ' Select second node (circle)
      opts = New PromptEntityOptions(vbLf & "Select end node (circle) for connection: ")
      opts.SetRejectMessage(vbLf & "That isn't a Circle!" & vbLf)
      opts.AddAllowedClass(GetType(Circle), True)
      Dim res2 As PromptEntityResult = ed.GetEntity(opts)
      If res2.Status <> PromptStatus.OK Then Exit Sub

      ' When creating links, we store a forward link (controlling information flow), but also a backlink. 
      ' The backlink is so a node that is moved can tell the node pointing to it to update it's graphics.
      ' (The graphics update is the only use we have for the back link list).
      ' Each link type is stored in a different Xrecord in the node's extension dictionary

      'Create resbuf with our link and add it to any existing links stored in our xrecord.
      Dim newResbuf As New ResultBuffer(New TypedValue(DxfCode.SoftPointerId, res2.ObjectId))
      Dim resbuf As ResultBuffer = SGPOverruleBag.GetXrecordData(res.ObjectId, SGPOverruleBag.LinkName)
      If resbuf Is Nothing Then
        resbuf = newResbuf
      Else
        resbuf = SGPOverruleBag.MergeResultBuffers(resbuf, newResbuf)
      End If
      SGPOverruleBag.SetXrecordData(res.ObjectId, resbuf, SGPOverruleBag.LinkName)

      'Now do the same with the back link
      newResbuf = New ResultBuffer(New TypedValue(DxfCode.SoftPointerId, res.ObjectId))
      resbuf = SGPOverruleBag.GetXrecordData(res2.ObjectId, SGPOverruleBag.BackLinkName)
      If resbuf Is Nothing Then
        resbuf = newResbuf
      Else
        resbuf = SGPOverruleBag.MergeResultBuffers(resbuf, newResbuf)
      End If
      SGPOverruleBag.SetXrecordData(res2.ObjectId, resbuf, SGPOverruleBag.BackLinkName)

      ' Regen our graphics to reflect updates
      ed.Regen()
    End Sub

    ' Removes link between two nodes
    <CommandMethod("REMOVELINK")> _
   Public Shared Sub RemoveLink()
      Dim db As Database = Application.DocumentManager.MdiActiveDocument.Database
      Dim ed As Editor = Application.DocumentManager.MdiActiveDocument.Editor
      ' Select first node
      Dim opts As New PromptEntityOptions(vbLf & "Select node (circle) to remove connection from: ")
      opts.SetRejectMessage(vbLf & "That isn't a Circle!" & vbLf)
      opts.AddAllowedClass(GetType(Autodesk.AutoCAD.DatabaseServices.Circle), True)
      Dim res As PromptEntityResult = ed.GetEntity(opts)

      If res.Status <> PromptStatus.OK Then Exit Sub

      'If this node doesn't have a Link xrecord, then we exit the command - there are no links to remove
      Dim resbuf As ResultBuffer = SGPOverruleBag.GetXrecordData(res.ObjectId, SGPOverruleBag.LinkName)
      If resbuf Is Nothing Then
        ed.WriteMessage(vbLf & "That node has no connections to remove!" & vbLf)
        Exit Sub
      End If

      ' Select second node
      opts = New PromptEntityOptions(vbLf & "Select connected node (circle) to remove: ")
      opts.SetRejectMessage(vbLf & "That isn't a Circle!" & vbLf)
      opts.AddAllowedClass(GetType(Autodesk.AutoCAD.DatabaseServices.Circle), True)
      Dim res2 As PromptEntityResult = ed.GetEntity(opts)

      If res2.Status <> PromptStatus.OK Then Exit Sub

      ' If this node has no backlinks then it can't be connected to the first selected node.
      ' If the second node does have backlink data then it may not be connected to our first node, 
      '  but that's ok - our RemoveResultBufferEntry helper function can cope with that.
      Dim resbuf2 As ResultBuffer = SGPOverruleBag.GetXrecordData(res2.ObjectId, SGPOverruleBag.BackLinkName)
      If resbuf2 Is Nothing Then
        ed.WriteMessage(vbLf & "That circle is not the receiver of any connections!" & vbLf)
        Exit Sub
      End If

      ' Remove link data from first node (has no effect if there is no link data with the specified objectid)
      Dim tv As New TypedValue(DxfCode.SoftPointerId, res2.ObjectId)
      resbuf = SGPOverruleBag.RemoveResultBufferEntry(resbuf, tv)
      SGPOverruleBag.SetXrecordData(res.ObjectId, resbuf, SGPOverruleBag.LinkName)

      ' Remove backlink data from second node (has no effect if there is no backlink data with the specified objectid)
      tv = New TypedValue(DxfCode.SoftPointerId, res.ObjectId)
      resbuf2 = SGPOverruleBag.RemoveResultBufferEntry(resbuf2, tv)
      SGPOverruleBag.SetXrecordData(res2.ObjectId, resbuf2, SGPOverruleBag.BackLinkName)

      ' Regen our graphics to reflect updates
      ed.Regen()

    End Sub

    ' Add any number of arbitrary text strings as information to a node. This text is displayed when the node is drawn.
    ' Each text string must be <= 127 chars.
    <CommandMethod("ADDINFO")> _
     Public Shared Sub AddInfo()
      Dim db As Database = Application.DocumentManager.MdiActiveDocument.Database
      Dim ed As Editor = Application.DocumentManager.MdiActiveDocument.Editor

      ' Select node (which doesn't have to be connected to anything else)
      Dim opts As New PromptEntityOptions(vbLf & "Select node to add information to: ")
      opts.SetRejectMessage(vbLf & "That isn't a Circle!" & vbLf)
      opts.AddAllowedClass(GetType(Autodesk.AutoCAD.DatabaseServices.Circle), True)
      Dim res As PromptEntityResult = ed.GetEntity(opts)

      If res.Status <> PromptStatus.OK Then Exit Sub
      'Prompt for the text to add as information
      Dim opts2 As New PromptStringOptions(vbLf & "Enter information (<128 characters):")
      opts2.AllowSpaces = True
      Dim res2 As PromptResult = ed.GetString(opts2)

      If res.Status <> PromptStatus.OK Then Exit Sub

      ' Ensure text string is 127 chars max.
      Dim len As Integer
      Dim str As String = res2.StringResult
      len = str.Length
      If len > 127 Then len = 127
      str = str.Substring(0, len)

      'Add text to info xrecord entry (won't allow duplicates)
      Dim newResbuf As New ResultBuffer(New TypedValue(DxfCode.XTextString, str))
      Dim resbuf As ResultBuffer = SGPOverruleBag.GetXrecordData(res.ObjectId, SGPOverruleBag.InfoName)
      If resbuf Is Nothing Then
        resbuf = newResbuf
      Else
        resbuf = SGPOverruleBag.MergeResultBuffers(resbuf, newResbuf)
      End If
      SGPOverruleBag.SetXrecordData(res.ObjectId, resbuf, SGPOverruleBag.InfoName)
      ' Regen our graphics to reflect updates
      ed.Regen()
    End Sub


    ' Remove all information from a seelcted node.
    <CommandMethod("REMOVEINFO")> _
 Public Shared Sub RemoveInfo()
      Dim db As Database = Application.DocumentManager.MdiActiveDocument.Database
      Dim ed As Editor = Application.DocumentManager.MdiActiveDocument.Editor
      'Select node
      Dim opts As New PromptEntityOptions(vbLf & "Select node to add information to: ")
      opts.SetRejectMessage(vbLf & "That isn't a Circle!" & vbLf)
      opts.AddAllowedClass(GetType(Autodesk.AutoCAD.DatabaseServices.Circle), True)
      Dim res As PromptEntityResult = ed.GetEntity(opts)

      If res.Status <> PromptStatus.OK Then Exit Sub

      ' Remove its data, if there's any data attached.
      Dim resbuf As ResultBuffer = SGPOverruleBag.GetXrecordData(res.ObjectId, SGPOverruleBag.InfoName)
      If Not resbuf Is Nothing Then
        SGPOverruleBag.SetXrecordData(res.ObjectId, New ResultBuffer, SGPOverruleBag.InfoName)
        ' Regen our graphics to reflect updates
        ed.Regen()
      End If

    End Sub


    ' Propagate information through network. Invoke this command to increment network by one timestep.
    <CommandMethod("STEPNETWORK")> _
    Public Shared Sub StepNetwork()
      ' This routine could be improved by maintaining a central list of ObjectIds of all circles in our network (i.e. use a 'manager' object).
      ' I'm just iterating the entire blocktablerecord contents for the current space and checking each circle to see if it has my dictionary in its extension dictionary.
      Dim db As Database = Application.DocumentManager.MdiActiveDocument.Database
      Dim ed As Editor = Application.DocumentManager.MdiActiveDocument.Editor

      ' This dictionary will store a list of nodes (ObjectIds) and the data that will be passed to them at the end of this timestep.
      Dim networkMap As New System.Collections.Generic.Dictionary(Of ObjectId, ResultBuffer)

      ' Iterate through all circles in the current space
      Using trans As Transaction = db.TransactionManager.StartTransaction

        ' Store list (in dictionary with objid as key) of circles in our network (that have fwd links), and what they point to.
        ' Create dictionary of nodes in our network. Accumulate  list of all incoming information for that node.
        ' (That's information being passed to that node by nodes pointing to it).
        Dim btr As BlockTableRecord = trans.GetObject(db.CurrentSpaceId, OpenMode.ForRead)

        For Each objId As ObjectId In btr
          'if its a circle, then we need to open it and check for our data
          If objId.ObjectClass.DxfName = "CIRCLE" Then
            Dim obj As DBObject = trans.GetObject(objId, OpenMode.ForRead)
            Dim linkResbuf As ResultBuffer = SGPOverruleBag.GetXrecordData(obj, SGPOverruleBag.LinkName)
            If Not linkResbuf Is Nothing Then
              'It has a forward link, so needs further processing.
              'First add this circle to the network map if it isn't already there.
              'We do this because we need to also include circles with no incoming data so we can remove information from them (as it flows out).
              If Not networkMap.ContainsKey(obj.ObjectId) Then
                networkMap.Add(obj.ObjectId, Nothing)
              End If
              'Now process each of the forward links from this node
              Dim arr() As TypedValue = linkResbuf.AsArray
              For Each tv As TypedValue In arr
                If tv.TypeCode = DxfCode.SoftPointerId Then ' Check value is correct type
                  Dim tmpObjId As ObjectId = tv.Value
                  'Fetch info for this node that will be passed to each node it has a forward link to.
                  Dim infoResBuf As ResultBuffer = SGPOverruleBag.GetXrecordData(obj, SGPOverruleBag.InfoName)
                  ' Add that information to the linked nodes.
                  If networkMap.ContainsKey(tmpObjId) Then
                    Dim oldResbuf As ResultBuffer = networkMap.Item(tmpObjId)
                    ' Now add to the dictionary the info that will be passed to the (forward) linked node from this current node.
                    networkMap.Item(tmpObjId) = SGPOverruleBag.MergeResultBuffers(oldResbuf, infoResBuf)
                  Else
                    ' If this objectid isn't in our map yet, then we add it along with the information resbuf from the node that was pointing to it.
                    networkMap.Add(tmpObjId, infoResBuf)
                  End If
                End If
              Next

            End If
          End If
        Next

        'Now we propagate the information to the new nodes
        For Each pair As KeyValuePair(Of ObjectId, ResultBuffer) In networkMap
          SGPOverruleBag.SetXrecordData(pair.Key, pair.Value, SGPOverruleBag.InfoName)
        Next

        trans.Commit()
      End Using
      ' Regen to update network display
      ed.Regen()
    End Sub


  End Class

#End Region




  Public Class SGPOverruleBag
    Implements IDisposable

#Region "Helper functions"
    Private Shared mRegisterClass As RXClass
    Private mOverrules As System.Collections.Generic.List(Of Overrule)

    ' Names used to store data in extension dictionary.
    Private Const mLinkName = "SGP_FWDLINKS"
    Private Const mBackLinkName = "SGP_BACKLINKS"
    Private Const mInfoName = "SGP_INFO"
    Private Const mDictName = "SGP_DATA"
    Friend Shared ReadOnly Property DictName()
      Get
        Return mDictName
      End Get
    End Property
    Friend Shared ReadOnly Property LinkName()
      Get
        Return mLinkName
      End Get
    End Property
    Friend Shared ReadOnly Property BackLinkName()
      Get
        Return mBackLinkName
      End Get
    End Property
    Friend Shared ReadOnly Property InfoName()
      Get
        Return mInfoName
      End Get
    End Property


    'Two versions of New - if no RXClass is passed in then we instantiate the overrules, but don't register them
    Public Sub New()
      subNew()
    End Sub


    Public Sub New(ByVal regClass As RXClass)
      subNew()
      RegisterClass = regClass
      If Not regClass Is Nothing Then
        Register(RegisterClass)
      End If
    End Sub


    Private Sub subNew()
      MyOverrules = New System.Collections.Generic.List(Of Overrule)
      MyOverrules.Add(New SGPDrawOverrule)
      MyOverrules.Add(New SGPObjectOverrule)
    End Sub


    ' Set the  class that overrules are rregistered for
    Friend Shared Property RegisterClass() As RXClass
      Get
        Return mRegisterClass
      End Get
      Set(ByVal value As RXClass)
        mRegisterClass = value
      End Set
    End Property


    ' List of overrules applied by DSOverruleBag
    Private Property MyOverrules() As System.Collections.Generic.List(Of Overrule)
      Get
        Return mOverrules
      End Get
      Set(ByVal value As System.Collections.Generic.List(Of Overrule))
        mOverrules = value
      End Set
    End Property


    ' Register and Unregister will normally be called from this class's New and Dispose methods.
    ' However, they may be called from another class to register the overrules against more than one class.
    Friend Sub Register(ByVal regClass As RXClass)
      If Not regClass Is Nothing Then
        For Each orule As Overrule In MyOverrules
          Overrule.AddOverrule(regClass, orule, False)
          orule.SetExtensionDictionaryEntryFilter(DictName)
        Next
        Overrule.Overruling = True
      End If
    End Sub
    Friend Sub Register()
      Register(RegisterClass)
    End Sub
    Friend Sub Unregister(ByVal regClass As RXClass)
      If Not regClass Is Nothing Then
        For Each orule As Overrule In MyOverrules
          Overrule.RemoveOverrule(regClass, orule)
        Next
        Overrule.Overruling = False
      End If
    End Sub
    Friend Sub Unregister()
      Unregister(RegisterClass)
    End Sub

    Friend Shared Function GetXrecordData(ByVal objId As ObjectId, ByVal entryName As String) As ResultBuffer
      Dim resbuf As ResultBuffer
      Dim db As Database = objId.Database
      Using trans As OpenCloseTransaction = db.TransactionManager.StartOpenCloseTransaction
        Dim obj As DBObject = trans.GetObject(objId, OpenMode.ForRead)
        resbuf = GetXrecordData(obj, entryName)
        trans.Commit()
      End Using
      Return resbuf
    End Function




    Friend Shared Function GetXrecordData(ByVal obj As DBObject, ByVal entryName As String) As ResultBuffer
      Dim resbuf As ResultBuffer
      Dim db As Database = obj.Database
      Dim nodId As ObjectId

      Using trans As OpenCloseTransaction = db.TransactionManager.StartOpenCloseTransaction

        nodId = obj.ExtensionDictionary
        If nodId = ObjectId.Null Then Return Nothing
        Dim nod As DBDictionary = trans.GetObject(nodId, OpenMode.ForRead)

        Dim dictId As ObjectId
        If Not nod.Contains(DictName) Then Return Nothing
        dictId = nod.GetAt(DictName)
        Dim dict As DBDictionary = trans.GetObject(dictId, OpenMode.ForRead)

        Dim xrecId As ObjectId
        If Not dict.Contains(entryName) Then Return Nothing
        xrecId = dict.GetAt(entryName)
        Dim xrec As Xrecord = trans.GetObject(xrecId, OpenMode.ForRead)
        resbuf = xrec.Data

        trans.Commit()
      End Using
      Return resbuf
    End Function




    Friend Shared Sub SetXrecordData(ByVal objId As ObjectId, ByVal resbuf As ResultBuffer, ByVal dictName As String)
      Dim db As Database = objId.Database
      Using trans As OpenCloseTransaction = db.TransactionManager.StartOpenCloseTransaction
        Try
          Dim obj As DBObject = trans.GetObject(objId, OpenMode.ForWrite)
          SetXrecordData(obj, resbuf, dictName)
          trans.Commit()
        Catch ex As Autodesk.AutoCAD.Runtime.Exception
          If ex.ErrorStatus <> ErrorStatus.WasOpenForWrite Then
            Throw (ex)
          End If
        End Try
      End Using
    End Sub



    Friend Shared Sub SetXrecordData(ByVal obj As DBObject, ByVal resbuf As ResultBuffer, ByVal entryName As String)
      Dim db As Database = obj.Database
      Dim extDictId As ObjectId
      Using trans As OpenCloseTransaction = db.TransactionManager.StartOpenCloseTransaction
        Try
          extDictId = obj.ExtensionDictionary
          If extDictId = ObjectId.Null Then
            obj.CreateExtensionDictionary()
            extDictId = obj.ExtensionDictionary
          End If
          Dim extDict As DBDictionary = trans.GetObject(extDictId, OpenMode.ForRead)
          Dim dictId As ObjectId
          If extDict.Contains(DictName) Then
            dictId = extDict.GetAt(DictName)
          Else
            Using tmpDict As New DBDictionary
              If Not extDict.IsWriteEnabled Then extDict.UpgradeOpen()
              dictId = extDict.SetAt(DictName, tmpDict)
              trans.AddNewlyCreatedDBObject(tmpDict, True)
            End Using
          End If
          Dim dict As DBDictionary = trans.GetObject(dictId, OpenMode.ForRead)
          Dim xrecId As ObjectId
          If dict.Contains(entryName) Then
            xrecId = dict.GetAt(entryName)
          Else
            Using tmpXrec As New Xrecord
              If Not dict.IsWriteEnabled Then dict.UpgradeOpen()
              dict.TreatElementsAsHard = True
              xrecId = dict.SetAt(entryName, tmpXrec)
              trans.AddNewlyCreatedDBObject(tmpXrec, True)
            End Using
          End If
          Dim xrec As Xrecord = trans.GetObject(xrecId, OpenMode.ForWrite)
          If resbuf Is Nothing Then resbuf = New ResultBuffer
          xrec.Data = resbuf
          trans.Commit()
        Catch ex As Autodesk.AutoCAD.Runtime.Exception
          If ex.ErrorStatus <> ErrorStatus.WasOpenForWrite Then
            Throw (ex)
          End If
        End Try
      End Using
    End Sub






    'Removes element tv from resbuf
    Friend Shared Function RemoveResultBufferEntry(ByVal resbuf As ResultBuffer, ByVal tv As TypedValue) As ResultBuffer
      Dim set1 As New HashSet(Of TypedValue)(resbuf.AsArray)
      If set1.Contains(tv) Then
        set1.Remove(tv)
        Dim newResbuf As New ResultBuffer
        ' Convert set to a resbuf
        For Each tvItem As TypedValue In set1
          ' Some operations result in our resbufs holding null ObjectIds.
          ' If we try to add a null ObjectID to the resbuf then an exception is thrown.
          ' This try...catch block quietly handles that case. This is a little bit ugly, but it works.
          Try
            newResbuf.Add(tvItem)
          Catch ex As Autodesk.AutoCAD.Runtime.Exception
            If ex.ErrorStatus <> ErrorStatus.NullObjectId Then Throw ex
          End Try
        Next
        Return newResbuf
      Else
        Return resbuf
      End If
    End Function



    'Merges two resbufs, making sure there are no duplicate entries (i.e. no two TypedValues [Key+Value] are the same)
    Friend Shared Function MergeResultBuffers(ByVal resbuf1 As ResultBuffer, ByVal resbuf2 As ResultBuffer) As ResultBuffer

      If resbuf1 Is Nothing Then Return resbuf2 ' Returned value can still be nothing
      If resbuf2 Is Nothing Then Return resbuf1 ' Ditto

      ' If we get to here, then both resbufs contain something
      Dim set1 As New HashSet(Of TypedValue)(resbuf1.AsArray)
      Dim set2 As New HashSet(Of TypedValue)(resbuf2.AsArray)
      set1.UnionWith(set2)

      Dim newResbuf As New ResultBuffer
      ' Convert list to a resbuf
      For Each tv As TypedValue In set1
        ' Some operations result in our resbufs holding null ObjectIds.
        ' If we try to add a null ObjectID to the resbuf then an exception is thrown.
        ' This try...catch block quietly handles that case. This is a little bit ugly, but it works.
        Try
          newResbuf.Add(tv)
        Catch ex As Autodesk.AutoCAD.Runtime.Exception
          If ex.ErrorStatus <> ErrorStatus.NullObjectId Then Throw ex
        End Try
      Next
      Return newResbuf

    End Function



    Private disposedValue As Boolean = False        ' To detect redundant calls

    ' IDisposable
    Protected Overridable Overloads Sub Dispose(ByVal disposing As Boolean)
      If Not Me.disposedValue Then
        If disposing Then
          ' TODO: free other state (managed objects).
          Unregister(RegisterClass)
          For Each orule As Overrule In MyOverrules
            orule.Dispose()
          Next
          mOverrules = Nothing
        End If
        ' TODO: free your own state (unmanaged objects).
        ' TODO: set large fields to null.
      End If
      Me.disposedValue = True
    End Sub


#Region " IDisposable Support "
    ' This code added by Visual Basic to correctly implement the disposable pattern.
    Public Overloads Sub Dispose() Implements IDisposable.Dispose
      ' Do not change this code.  Put cleanup code in Dispose(ByVal disposing As Boolean) above.
      Dispose(True)
      GC.SuppressFinalize(Me)
    End Sub

    Protected Overrides Sub Finalize()
      Dispose(False)
      MyBase.Finalize()
    End Sub

#End Region

#End Region

#Region "Overrules"

    Private Class SGPObjectOverrule
      Inherits ObjectOverrule

      ' If this node is moved then we want to tell the node pointing to it to updates its graphical pointer.
      ' Doing this in the Close overrule covers grip editing and also editing via the Property Palette (or an API call)
      ' It would seem more sensible to overrule TransformBy, but that is not called when grip editing the circle center 
      '  or editing center via property palette. Doing it this way means our code is called more often, but it does get the job done.
      Public Overrides Sub Close(ByVal dbObject As Autodesk.AutoCAD.DatabaseServices.DBObject)

        If Not dbObject.Database Is Nothing And Not dbObject.IsUndoing And dbObject.IsModified Then
          Dim resbuf As ResultBuffer = GetXrecordData(dbObject, BackLinkName)
          If Not resbuf Is Nothing Then
            Dim arr() As TypedValue = resbuf.AsArray
            For Each tv As TypedValue In arr
              If tv.TypeCode = DxfCode.SoftPointerId Then
                Dim objId As ObjectId = tv.Value
                If objId.IsValid Then
                  ' This next part isn't ideal, but its risky to use transactions here.
                  ' We have to rely on silently catching an exception because we can't check if the object is open before we try to open it
                  ' The exception also  breaks us out of a possible infinite notification loop.
                  Try
                    ' Sometimes (during cloning) this linked object is a DBImpObject instead of a Circle
                    ' We check the object type here to avoid a preventable error
                    If objId.ObjectClass.DxfName = "CIRCLE" Then
                      Dim ent As Entity = objId.Open(OpenMode.ForWrite)
                      ent.RecordGraphicsModified(True)
                      ent.Close()
                    End If
                  Catch ex As Autodesk.AutoCAD.Runtime.Exception
                    If ex.ErrorStatus <> ErrorStatus.WasOpenForWrite Then
                      Throw (ex)
                    End If
                  End Try
                End If
              End If
            Next
          End If
        End If

        ' Call base Close function
        MyBase.Close(dbObject)

      End Sub


      ' When a node is erased, we have to update the forward and back pointers referencing it.
      Public Overrides Sub [Erase](ByVal dbObject As Autodesk.AutoCAD.DatabaseServices.DBObject, ByVal erasing As Boolean)
        Dim backLinkResbuf As ResultBuffer
        Dim fwdLinkResbuf As ResultBuffer
        If erasing Then
          ' First iterate through backlink list, removing fwd link pointers from those nodes
          backLinkResbuf = GetXrecordData(dbObject, BackLinkName)
          If Not backLinkResbuf Is Nothing Then
            Dim arr() As TypedValue = backLinkResbuf.AsArray
            For Each tv As TypedValue In arr
              If tv.TypeCode = DxfCode.SoftPointerId Then
                Dim objId As ObjectId = tv.Value
                If objId.IsValid Then
                  fwdLinkResbuf = GetXrecordData(objId, LinkName)
                  Dim tvToRemove As New TypedValue(DxfCode.SoftPointerId, dbObject.ObjectId)
                  Dim newResbuf As ResultBuffer = RemoveResultBufferEntry(fwdLinkResbuf, tvToRemove)
                  SetXrecordData(objId, newResbuf, LinkName)
                  'Tell entity to redraw itself (link arrow needs to be removed)
                  Try
                    Dim ent As Entity = objId.Open(OpenMode.ForWrite)
                    ent.RecordGraphicsModified(True)
                    ent.Close()
                  Catch ex As Autodesk.AutoCAD.Runtime.Exception
                    If ex.ErrorStatus <> ErrorStatus.WasOpenForWrite Then
                      Throw (ex)
                    End If
                  End Try
                End If
              End If
            Next
          End If

          'Then iterate through fwd link list, removing back pointers from those
          fwdLinkResbuf = GetXrecordData(dbObject, LinkName)
          If Not fwdLinkResbuf Is Nothing Then
            Dim arr() As TypedValue = fwdLinkResbuf.AsArray
            For Each tv As TypedValue In arr
              If tv.TypeCode = DxfCode.SoftPointerId Then
                Dim objId As ObjectId = tv.Value
                If objId.IsValid Then
                  backLinkResbuf = GetXrecordData(objId, BackLinkName)
                  Dim tvToRemove As New TypedValue(DxfCode.SoftPointerId, dbObject.ObjectId)
                  Dim newResbuf As ResultBuffer = RemoveResultBufferEntry(backLinkResbuf, tvToRemove)
                  SetXrecordData(objId, newResbuf, BackLinkName)
                End If
              End If
            Next
          End If
        Else
          ' If we're unerasing, then all we have to do is tell nodes pointing to this one to redraw their pointers.
          ' So we just iterate through back link list calling RecordGraphicsModified. 
          backLinkResbuf = GetXrecordData(dbObject, BackLinkName)
          If Not backLinkResbuf Is Nothing Then
            Dim arr() As TypedValue = backLinkResbuf.AsArray
            For Each tv As TypedValue In arr
              If tv.TypeCode = DxfCode.SoftPointerId Then
                Dim objId As ObjectId = tv.Value
                If objId.IsValid Then
                  'Tell entity to redraw itself (link arrow needs to be redrawn)
                  Try
                    Dim ent As Entity = objId.Open(OpenMode.ForWrite)
                    ent.RecordGraphicsModified(True)
                    ent.Close()
                  Catch ex As Autodesk.AutoCAD.Runtime.Exception
                    If ex.ErrorStatus <> ErrorStatus.WasOpenForWrite Then
                      Throw (ex)
                    End If
                  End Try

                End If
              End If
            Next
          End If
        End If

        ' Call base implementation
        MyBase.[Erase](dbObject, erasing)
      End Sub


      ' We deliberately only clone nodes in the network downstream of the selected node(s). I.e. We use forward pointer list, but not back pointer list.
      Public Overrides Function DeepClone(ByVal dbObject As Autodesk.AutoCAD.DatabaseServices.DBObject, ByVal ownerObject As Autodesk.AutoCAD.DatabaseServices.DBObject, ByVal idMap As Autodesk.AutoCAD.DatabaseServices.IdMapping, ByVal isPrimary As Boolean) As Autodesk.AutoCAD.DatabaseServices.DBObject
        Dim resbuf As ResultBuffer
        Dim clonedObj As DBObject = Nothing

        Try
          If idMap.Contains(dbObject.ObjectId) Then
            clonedObj = Nothing
          Else
            ' Clone this node
            clonedObj = MyBase.DeepClone(dbObject, ownerObject, idMap, isPrimary)
          End If

          ' Clone nodes listed in link data
          ' Check on isPrimary ensures we don't go into an infinite loop.
          If isPrimary Then
            resbuf = GetXrecordData(dbObject, LinkName)
            If Not resbuf Is Nothing Then
              Dim arr() As TypedValue = resbuf.AsArray
              For Each tv As TypedValue In arr
                If tv.TypeCode = DxfCode.SoftPointerId Then
                  Dim objId As ObjectId = tv.Value
                  If objId.IsValid Then
                    'if we don't find the ObjId of our linked entity in the idMap, then we must clone it it
                    If Not idMap.Contains(objId) Then
                      If objId.ObjectClass.DxfName = "CIRCLE" Then 'Have to include this check because object can sometimes become ImpDbObject during cloning
                        Using trans As OpenCloseTransaction = dbObject.Database.TransactionManager.StartOpenCloseTransaction
                          ' Open linked object using ObjectId retrieved from Xrecord
                          Dim linkedObj As DBObject = trans.GetObject(objId, OpenMode.ForRead)
                          ' Add linked object to deepclone operation
                          Dim linkedObjClone As DBObject = linkedObj.DeepClone(ownerObject, idMap, isPrimary)
                          trans.AddNewlyCreatedDBObject(linkedObjClone, True)
                          trans.Commit()
                        End Using
                      End If
                    End If
                  End If
                End If
              Next
            End If

          End If
        Catch ex As Autodesk.AutoCAD.Runtime.Exception
          ' Silently catch any AutoCAD exceptions and continue. 
          ' (We assume any exceptions thrown from the code are non-fatal and simply resulted in the cloning not working properly).

        End Try
        Return clonedObj
      End Function

      ' Overruling WBlockClone is the reason we use softpointers. Hardpointers automatically bring objects into the cloning operation.
      ' That causes problems with circular dependencies, which results in the linked nodes not being cloned.
      ' We overcome that by using softpoints, and adding the linked entities to the cloning operation ourselves.
      ' We deliberately only clone nodes in the network downstream of the selected node(s). I.e. We use forward pointer list, but not back pointer list.
      Public Overrides Function WblockClone(ByVal dbObject As Autodesk.AutoCAD.DatabaseServices.DBObject, ByVal ownerObject As Autodesk.AutoCAD.Runtime.RXObject, ByVal idMap As Autodesk.AutoCAD.DatabaseServices.IdMapping, ByVal isPrimary As Boolean) As Autodesk.AutoCAD.DatabaseServices.DBObject
        Dim resbuf As ResultBuffer
        Dim clonedObj As DBObject = Nothing

        Try
          If idMap.Contains(dbObject.ObjectId) Then
            Return Nothing
          Else
            ' Clone this node and make sure its owner is set correctly
            clonedObj = MyBase.WblockClone(dbObject, ownerObject, idMap, isPrimary)
            If clonedObj.OwnerId = ObjectId.Null Then
              clonedObj.OwnerId = idMap.Lookup(dbObject.OwnerId).Value
            End If

          End If

          ' Check on isPrimary ensures we don't go into an infinite loop.
          If isPrimary Then
            'Clone nodes listed in link data
            resbuf = GetXrecordData(dbObject, LinkName)
            If Not resbuf Is Nothing Then
              Dim arr() As TypedValue = resbuf.AsArray
              For Each tv As TypedValue In arr
                If tv.TypeCode = DxfCode.SoftPointerId Then
                  Dim objid As ObjectId = tv.Value
                  If objid.IsValid Then
                    'if we don't find the objId of our linked entity in the idMap, then we must clone it it
                    If Not idMap.Contains(objid) Then
                      Using trans As OpenCloseTransaction = dbObject.Database.TransactionManager.StartOpenCloseTransaction
                        ' Open linked entity
                        Dim linkedObj As DBObject = trans.GetObject(objid, OpenMode.ForRead)
                        ' Add linked entity to the wblockclone operaiton
                        Dim linkedObjClone As DBObject = linkedObj.WblockClone(ownerObject, idMap, isPrimary)
                        trans.AddNewlyCreatedDBObject(linkedObjClone, True)
                        trans.Commit()
                      End Using
                    End If
                  End If
                End If
              Next
            End If
          End If

        Catch ex As Autodesk.AutoCAD.Runtime.Exception
          ' Silently catch any AutoCAD exceptions and continue. 
          ' (We assume any exceptions thrown from the code are non-fatal and simply resulted in the cloning not working properly).

        End Try

        Return clonedObj

      End Function

    End Class



    Private Class SGPDrawOverrule
      Inherits DrawableOverrule

      Public Overrides Function WorldDraw(ByVal drawable As Autodesk.AutoCAD.GraphicsInterface.Drawable, ByVal wd As Autodesk.AutoCAD.GraphicsInterface.WorldDraw) As Boolean

        Dim myCircle As Circle = drawable
        Dim db As Database = myCircle.Database

        If Not db Is Nothing Then
          ' Draw arrow to all nodes this node has forward links to
          Using trans As OpenCloseTransaction = db.TransactionManager.StartOpenCloseTransaction
            Using rb As ResultBuffer = GetXrecordData(myCircle, LinkName)
              If Not rb Is Nothing Then
                Dim arr() As TypedValue = rb.AsArray
                For Each tv As TypedValue In arr
                  If tv.TypeCode = DxfCode.SoftPointerId Then
                    Dim objId As ObjectId = tv.Value
                    If objId.IsValid Then
                      Dim circle2 As Circle = trans.GetObject(objId, OpenMode.ForRead)
                      Dim pts As New Point3dCollection
                      pts.Add(myCircle.Center)
                      pts.Add(circle2.Center)
                      Dim vec As Vector3d = pts(1) - pts(0)
                      Dim perpVec As Vector3d = vec.CrossProduct(myCircle.Normal)
                      Dim pt2 As Point3d = circle2.Center - vec / 10 + perpVec / 30
                      Dim pt3 As Point3d = circle2.Center - vec / 10 - perpVec / 30
                      pts.Add(pt2)
                      pts.Add(pt3)
                      pts.Add(circle2.Center)
                      wd.Geometry.Polyline(pts, myCircle.Normal, New IntPtr(5321))
                    End If
                  End If
                Next

              End If
            End Using

            ' If this node contains information, then color it green and print the information
            Using rb As ResultBuffer = GetXrecordData(myCircle, InfoName)
              If Not rb Is Nothing Then
                wd.SubEntityTraits.Color = 3
                Dim arr() As TypedValue = rb.AsArray
                Dim size As Double
                ' Set text size so it scales to not extend below bottom of circle
                If arr.Count < 5 Then
                  size = myCircle.Radius / 5
                Else
                  size = myCircle.Radius / arr.Count
                End If

                ' Convert resbuf data to text on screen
                For index As Integer = 0 To arr.Count - 1
                  Dim msg As String = arr(index).Value.ToString
                  Dim txtVec As New Vector3d(0, size, 0)
                  wd.Geometry.Text(myCircle.Center - txtVec * index, myCircle.Normal, Vector3d.XAxis, size * 0.9, 1, 0, msg)
                Next

              End If
            End Using
            trans.Commit()
          End Using
        End If
        Return MyBase.WorldDraw(drawable, wd)

      End Function

    End Class

#End Region


  End Class


End Namespace